import time
from typing import Any, Dict, List

import numpy as np
from badger import environment
from epics import caget_many, caput, caget
import epics

from pydantic import ConfigDict, model_validator

from AWAControl import MAGNET_CONFIG, RF_CONFIG
from AWAControl.diagnostics.screen import load_screen, AWAImageDiagnostic

# get ranges and pvs for control names
ranges = {name: val["RANGE"] for name, val in (MAGNET_CONFIG | RF_CONFIG).items()}
pvs = {name: val["PV"] for name, val in (MAGNET_CONFIG | RF_CONFIG).items()}

screen_observables = [
        "Sx",
        "Sy",
        "Cx",
        "Cy",
        "total_size",
        "bb_penalty",
        "log10_total_intensity",
        "center_dist"
]

class Environment(environment.Environment):
    name = "awa"
    variables = ranges

    observables = [
        "Sx",
        "Sy",
        "Cx",
        "Cy",
        "total_size",
        "bb_penalty",
        "log10_total_intensity",
        "K1_phase_readback",
        "ICT1_nC"
    ]
    screen_name: str = None
    image_diagnostic: Any = None
    setpoints: dict = {}
    n_shots: int = 1

    model_config = ConfigDict(validate_assignment=True)

    @model_validator(mode="after")
    def validate_screen_name(self):
        if self.screen_name is not None:
            if self.image_diagnostic is None:
                self.image_diagnostic = load_screen(self.screen_name)
            elif isinstance(self.image_diagnostic, AWAImageDiagnostic):
                if self.image_diagnostic.alias != self.screen_name:
                    self.image_diagnostic = load_screen(self.screen_name)
            elif isinstance(self.image_diagnostic, dict):
                self.image_diagnostic = load_screen(self.screen_name)

        return self

    def get_observables(self, observable_names: List[str]) -> Dict:
        results = {}
        if "K1_phase_readback" in observable_names:
            results["K1_phase_readback"] = caget("AWALLRF:K1:Phase")

        if len(set(observable_names).intersection(set(screen_observables))):
            metadata = self.setpoints or {}
            metadata = metadata | results
            results.update(self.image_diagnostic.measure_beamsize(
                self.n_shots, **metadata
            ))
            sigma_xy = np.sqrt(np.array(results["Sx"]) ** 2 + np.array(results["Sy"]) ** 2)
            roundness = np.abs(np.array(results["Sx"]) - np.array(results["Sy"]))
            results["sigma_xy"] = sigma_xy
            results["roundness"] = roundness
            results["total_size"] = sigma_xy + roundness
        

            # measure distance to image center
            results["center_dist_x"] = np.array(results["Cx"]) - self.image_diagnostic.roi.xwidth / 2
            results["center_dist_y"] = np.array(results["Cy"]) - self.image_diagnostic.roi.ywidth / 2
    
            results["center_dist"] = (
                results["center_dist_x"] ** 2 + results["center_dist_y"] ** 2
            ) ** 0.5

        if "ICT1_nC" in observable_names:
            results["ICT1_nC"] = caget("AWAVXI11ICT:Ch1") * 1e9

        results["time"] = time.time()
        return results

    def get_variables(self, variable_names: List[str]) -> Dict:
        #epics.ca.clear_cache()
        results = caget_many([pvs[name] for name in variable_names])
        return dict(zip(variable_names, results))

    def set_variables(self, variable_inputs: Dict[str, float]):
        #epics.ca.clear_cache()
        if self.setpoints is not None:
            self.setpoints.update(variable_inputs)
        else:
            self.setpoints = variable_inputs
        for name, val in variable_inputs.items():
            caput(pvs[name], val)
