import time
from typing import List, Optional
from pydantic import BaseModel
from epics import pv
import pvaccess as pva


class AWADataAcq(BaseModel):
    """
    Defines an interface class that allows for fast, synchronized, data collection
    of camera images from multiple locations.
    """
    camera_pv_names: Optional[List[str]] = ["AWANIFG:ImgData"]
    ict_charge_pv_names: Optional[List[str]] = [
        "AWAVXI11ICT:Ch1",
        "AWAVXI11ICT:Ch2",
        "AWAVXI11ICT:Ch3",
        "AWAVXI11ICT:Ch4"
    ]
    ict_waveform_pv_names: Optional[List[str]] = [
        "AWAVXI11ICT:wf1",
        "AWAVXI11ICT:wf2",
        "AWAVXI11ICT:wf3",
        "AWAVXI11ICT:wf4"
    ]
    trigger_control_channel_number = 9
    include_ict_waveforms: bool = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.pv_objects = {}
        self.pva_channels = {}
        self._total_shots = 0
        self._done = False
        self._data = {}

        # create pvs
        self.create_pv_objects()

        # initialize pva channels
        self.initialize_pva_channels()

    @property
    def trigger_control_pv(self):
        return pv.PV("AWATrigCtrl:SetCh%d" % self.trigger_control_channel_number)

    @property
    def trigger_control_enable_pv(self):
        return pv.PV("AWATrigCtrl:EnableCh%d" % self.trigger_control_channel_number)

    def stop_trigger(self):
        # enable the trigger ctrl to stop trigger
        # make sure the master trigger was properly set
        self.trigger_control_enable_pv.put(1, wait=True)

        # Set the TrigCtrl input to 0 to stop the trigger in case master is
        # not active
        self.trigger_control_pv.put(0, wait=True)

    def start_trigger(self):
        self.trigger_control_pv.put(1, wait=True)
        self.trigger_control_enable_pv.put(0, wait=True)

    @property
    def all_pv_names(self):
        """ return all of the pv names that we want to monitor"""
        names = self.camera_pv_names
        if self.include_ict_charge_values:
            names += self.ict_charge_pv_names
        if self.include_ict_waveforms:
            names += self.ict_waveform_pv_names

        return names

    def create_pv_objects(self):
        """ create pv objects from camera_pv_list """
        self.pv_objects = {
            name: pv.PV(name, auto_monitor=False) for name in self.all_pv_names
        }

    def initialize_pva_channels(self):
        """ initialize pva channels for parallel evaluation """
        self.pva_channels = {
            name: pva.Channel(name) for name in self.all_pv_names
        }

    def measure(self, pvname, **kwargs):
        pv_object = self.pv_objects[pvname]
        data = pv_object.get(use_monitor=False)
        timestamp = pv_object.timestamp
        self._data[pvname].append(data)
        self._data[pvname+"_ts"].append(timestamp)

    def acquire(self, n_shots=50):
        self._data = {name: [] for name in list(self.pva_channels)}

        # subscribe processes
        for name, channel in self.pva_channels.items():
            channel.subscribe(name, self.measure)

        # reset trigger
        self.stop_trigger()

        # start monitors
        for name, channel in self.pva_channels.items():
            channel.startMonitor()

        time.sleep(2)

        # start the trigger
        self.start_trigger()

        # wait for data to accumulate
        while len(self._data[self.camera_pv_names[0]]) < n_shots:
            time.sleep(1)

        # stop monitors
        for _, channel in self.pva_channels.items():
            channel.stopMonitor()

        # unsubscribe processes
        for name, channel in self.pva_channels.items():
            channel.unsubscribe(name)

        # return data
        return self._data











