import time

from xopt import VOCS, Evaluator, Xopt
from xopt.generators import get_generator
from xopt.generators.bayesian.models.standard import StandardModelConstructor
from xopt.utils import get_local_region

from AWAControl import MAGNET_CONFIG
import matplotlib.pyplot as plt


def center_beam_on_screen(
        awa_env,
        screen_name,
        trim_name,
        start_at_zero=True,
        tolerance=5.0
):
    # start time to measbounding_box_half_widthure execution time
    start = time.time()

    # get information about trim magnets and screen settings from config files
    variable_names = [f"{trim_name}{ele}" for ele in ["H", "V"]]
    trim_info = MAGNET_CONFIG

    # set_screen
    awa_env.screen_name = screen_name

    # test measurement
    awa_env.image_diagnostic.test_measurement()
    plt.show()

    # awa_env.image_diagnostic.visualize = True

    # define evaluate function
    def evaluate(inputs: dict):
        # caput values
        awa_env.set_variables(inputs)

        # wait for changes to occur - use small wait time for interpolated measurements
        time.sleep(1.5)

        results = awa_env.get_observables(["center_dist"])
        # plt.show()

        return results

    # create Xopt objects
    # create variables, objectives, etc. based on config files
    vocs = VOCS(
        variables={
            name: awa_env.variables[name] for name in variable_names
        },
        objectives={"center_dist": "MINIMIZE"}
    )

    # create evaluator object
    evaluator = Evaluator(function=evaluate)

    # create generator object
    model_constructor = StandardModelConstructor(use_low_noise_prior=False)
    generator = get_generator("expected_improvement")(
        vocs=vocs, gp_constructor=model_constructor,
        n_interpolate_points=2,
        turbo_controller="optimize",
    )
    generator.numerical_optimizer.max_time = 10.0

    # create Xopt object
    X = Xopt(vocs=vocs, generator=generator, evaluator=evaluator, strict=True)
    # print(X)

    # initial random sampling

    # start samples at zero trim strength
    if start_at_zero:
        X.evaluate_data({name: 0.0 for name in X.vocs.variable_names})

    # get current PV values for trims
    current_value = awa_env.get_variables(X.vocs.variable_names)
    print(current_value)
    # get small region around current point to sample
    random_sample_region = get_local_region(current_value, X.vocs, fraction=0.25)
    print(random_sample_region)

    # do random sampling
    X.random_evaluate(4, custom_bounds=random_sample_region)

    # do optimization - stop when tolerance is reached
    for i in range(20):
        X.step()
        idx, val, _ = X.vocs.select_best(X.data)
        if abs(val) < tolerance:
            print(
                f"step {i}, best val {val}, current_val {float(X.data.iloc[-1]['center_dist'])}")
            break

        print(
            f"step {i}, best val {val}, current_val {float(X.data.iloc[-1]['center_dist'])}")

    # set trims to optimal config
    idx, val, _ = X.vocs.select_best(X.data)
    X.evaluate_data(X.data.iloc[idx][X.vocs.variable_names])

    # print the time it took to run
    print(time.time() - start)

    return X
