import time
import numpy as np

import torch
from xopt import VOCS, Xopt, Evaluator
from xopt.generators.bayesian.bax_generator import BaxGenerator
from AWAControl.algorithms.curv import CurvatureGridOptimize
from epics import caget, caput
from gpytorch.priors import GammaPrior
from xopt.generators.bayesian.turbo import EntropyTurboController
import matplotlib.pyplot as plt

import warnings
warnings.filterwarnings("ignore")

def run_schottky_scan(
    awa_env, n_steps=10
):
    # LLRF feedback old state
    pv = "AWALLRF:K1:EnableFC"
    old_llrf_feedback_state = caget(pv)

    # turn feedback off
    print("turning llrf feedback off")
    caput(pv, 0)
    
    
    def evaluate(inputs: dict):
        # caput values
        awa_env.set_variables(inputs)

        # wait for changes to occur - use small wait time for interpolated measurements
        time.sleep(1)

        results = awa_env.get_observables(['ICT1_nC'])
        print(results)
        return results

    vocs = VOCS(
        variables={
            'K1SetPhase': awa_env.variables['K1SetPhase']
        },
        observables=['ICT1_nC']
    )

    algorithm = CurvatureGridOptimize(
        n_mesh_points=200, minimize=False, 
        observable_names_ordered=["ICT1_nC"], 
        use_mean=True
    )

    # construct BAX generator
    generator = BaxGenerator(
        vocs=vocs, algorithm=algorithm,
        n_interpolate_points=5, 
        #turbo_controller=EntropyTurboController(vocs)
    )
    generator.gp_constructor.custom_noise_prior = GammaPrior(1.202, 8.744)
    #generator.turbo_controller.restrict_model_data = False

    X = Xopt(
        vocs=vocs,
        evaluator=Evaluator(function=evaluate),
        generator=generator,
    )

    # test measurement
    try:
        test_meas = awa_env.get_observables(['ICT1_nC'])
    except Exception as e:
        print("test measurement failed see error below")
        print(e)
        
    
    X.evaluate_data({'K1SetPhase': np.linspace(-20,360,20)})

    test_x = torch.linspace(*X.vocs.variables["K1SetPhase"], 200)

    for ii in range(n_steps):
        print(f"Iteration: {ii}")
        model = X.generator.train_model()

        # get acquisition function from generator
        acq = X.generator.get_acquisition(model)

        # calculate model posterior and acquisition function at each test point
        # NOTE: need to add a dimension to the input tensor for evaluating the
        # posterior and another for the acquisition function, see
        # https://botorch.org/docs/batching for details
        # NOTE: we use the `torch.no_grad()` environment to speed up computation by
        # skipping calculations for backpropagation
        with torch.no_grad():
            posterior = model.posterior(test_x.unsqueeze(1))
            acq_val = acq(test_x.reshape(-1, 1, 1))

        # get mean function and confidence regions
        mean = posterior.mean
        l, u = posterior.mvn.confidence_region()

        # plot model and acquisition function
        fig, ax = plt.subplots(3, 1, sharex="all")
        fig.set_size_inches(8, 6)

        # plot model posterior
        ax[0].plot(test_x, mean, label="Posterior mean")
        ax[0].fill_between(test_x, l, u, alpha=0.25,
                           label="Posterior confidence region")

        # add data to model plot
        ax[0].plot(X.data["K1SetPhase"], X.data["ICT1_nC"], "C1o",
                   label="Training data")

        if ii == 0:
            ax[0].legend(ncols=2)

        # plot the function samples and their optima found by BAX
        test_points = X.generator.algorithm_results["test_points"]
        posterior_samples = X.generator.algorithm_results["posterior_samples"]
        execution_paths = X.generator.algorithm_results["execution_paths"]

        label1 = 'Function Samples'
        label2 = 'Sample Optima'
        for i in range(len(posterior_samples)):
            ax[1].plot(test_points, posterior_samples[i], c='C0', alpha=0.3,
                                  label=label1)
            ax[1].scatter(*execution_paths[i], c='r', marker='x', s=80,
                                   label=label2, zorder=10)
            label1 = None
            label2 = None

        # add legend
        if ii == 0:
            ax[1].legend()

        # plot acquisition function
        ax[2].plot(test_x, acq_val.flatten())

        ax[0].set_ylabel("ICT1_nC")
        ax[1].set_ylabel("Virtual objective")
        ax[2].set_ylabel(r"$\alpha(x)$")
        ax[2].set_xlabel("K1SetPhase")

        mean = float(execution_paths[:, 0].mean())
        std = float(execution_paths[:, 0].std())

        print(f"Predicted zero phase location {mean} +/- {std}")
        X.step()
        
    print("Setting K1Phase to "+str(mean+50))
    evaluate({'K1SetPhase': mean+50})

    print("returning LLRF feedback from old state")
    caput(pv, old_llrf_feedback_state)
    
    return X
